from electrum.i18n import _

fullname = _('Revealer Backup Utility')
description = ''.join(["<br/>",
    "<b>"+_("Do you have something to hide ?")+"</b>", '<br/>', '<br/>',
    _("This plug-in allows you to create a visually encrypted backup of your wallet seeds, or of custom alphanumeric secrets."), '<br/>', '<br/>',
    _("For more information, visit"),
    " <a href=\"https://revealer.cc\">https://revealer.cc</a>", '<br/>', '<br/>',
])
available_for = ['qt']


